/****************************************************************************
**
** Copyright (C) 2012 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** You may use this file under the terms of the BSD license as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
**     the names of its contributors may be used to endorse or promote
**     products derived from this software without specific prior written
**     permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QGLWidget>
#include <QMatrix4x4>
#include <QVector3D>

#include <qmath.h>

#include "qtlogo.h"

static const qreal tee_height = 0.311126;
static const qreal cross_width = 0.25;
static const qreal bar_thickness = 0.113137;
static const qreal inside_diam = 0.20;
static const qreal outside_diam = 0.30;
static const qreal logo_depth = 0.10;
static const int num_divisions = 32;

//! [0]
struct Geometry
{
    QVector<GLushort> faces;
    QVector<QVector3D> vertices;
    QVector<QVector3D> normals;
    void appendSmooth(const QVector3D &a, const QVector3D &n, int from);
    void appendFaceted(const QVector3D &a, const QVector3D &n);
    void finalize();
    void loadArrays() const;
};
//! [0]

using namespace std;
using namespace giac;

bool find_zscale(const gen & surface,double & zmin,double & zmax){
  if (surface.type!=_VECT)
    return true;
  if (surface.subtype!=_POLYEDRE__VECT && surface.subtype!=_GROUP__VECT && surface._VECTptr->size()==3){
    if (!is_zero(re(surface._VECTptr->back(),context0)))
  return false;
    gen s3=evalf_double(surface._VECTptr->back()/cst_i,2,context0);
    if (s3.type==_DOUBLE_){
  double s3d = s3._DOUBLE_val;
  if (s3d<zmin) zmin=s3d;
  if (s3d>zmax) zmax=s3d;
    }
    return true;
  }
  const_iterateur it=surface._VECTptr->begin(),itend=surface._VECTptr->end();
  for (;it!=itend;++it){
    if (!find_zscale(*it,zmin,zmax))
  return false;
  }
  return true;
}


bool get_glvertex(const vecteur & v,double & d1,double & d2,double & d3,double realiscmplx,double zmin,GIAC_CONTEXT){
  if (v.size()==3){
    gen tmp;
    tmp=evalf_double(v[0],2,contextptr);
    if (tmp.type!=_DOUBLE_) return false;
    d1=tmp._DOUBLE_val;
    tmp=evalf_double(v[1],2,contextptr);
    if (tmp.type!=_DOUBLE_) return false;
    d2=tmp._DOUBLE_val;
    tmp=evalf_double(v[2],2,contextptr);
    if (realiscmplx){
      double arg=0;
      if (realiscmplx<0){
        d3=evalf_double(im(tmp,contextptr),2,contextptr)._DOUBLE_val-zmin;
        arg=-d3*realiscmplx;
      }
      else {
        if (tmp.type==_DOUBLE_){
          d3=tmp._DOUBLE_val;
          if (d3<0){
            arg=M_PI;
            d3=-d3;
          }
        }
        else {
          if (tmp.type==_CPLX && tmp._CPLXptr->type==_DOUBLE_ && (tmp._CPLXptr+1)->type==_DOUBLE_){
            double r=tmp._CPLXptr->_DOUBLE_val;
            double i=(tmp._CPLXptr+1)->_DOUBLE_val;
            arg=std::atan2(i,r);
            d3=std::sqrt(r*r+i*i);
          }
          else
            return false;
        }
      } // end realiscmplx>0
        // set color corresponding to argument
    }
    else {
      if (tmp.type!=_DOUBLE_) return false;
      d3=tmp._DOUBLE_val;
    }
    return true;
  }
  return false;
}


bool get_glvertex(const gen & g,double & d1,double & d2,double & d3,double realiscmplx,double zmin,GIAC_CONTEXT){
  if (g.type!=_VECT)
    return false;
  return get_glvertex(*g._VECTptr,d1,d2,d3,realiscmplx,zmin,contextptr);
}


//! [1]
class Patch
{
public:
    enum Smoothing { Faceted, Smooth };
    Patch(Geometry *);
    void setSmoothing(Smoothing s) { sm = s; }
    void translate(const QVector3D &t);
    void rotate(qreal deg, QVector3D axis);
    void draw() const;
    void addTri(const QVector3D &a, const QVector3D &b, const QVector3D &c, const QVector3D &n);
    void addQuad(const QVector3D &a, const QVector3D &b,  const QVector3D &c, const QVector3D &d);

    GLushort start;
    GLushort count;
    GLushort initv;

    GLfloat faceColor[4];
    QMatrix4x4 mat;
    Smoothing sm;
    Geometry *geom;
};
//! [1]

static inline void qSetColor(float colorVec[], QColor c)
{
    colorVec[0] = c.redF();
    colorVec[1] = c.greenF();
    colorVec[2] = c.blueF();
    colorVec[3] = c.alphaF();
}

void Geometry::loadArrays() const
{
    glVertexPointer(3, GL_FLOAT, 0, vertices.constData());
    glNormalPointer(GL_FLOAT, 0, normals.constData());
}

void Geometry::finalize()
{
    // TODO: add vertex buffer uploading here

    // Finish smoothing normals by ensuring accumulated normals are returned
    // to length 1.0.
    for (int i = 0; i < normals.count(); ++i)
        normals[i].normalize();
}

void Geometry::appendSmooth(const QVector3D &a, const QVector3D &n, int from)
{
    // Smooth normals are acheived by averaging the normals for faces meeting
    // at a point.  First find the point in geometry already generated
    // (working backwards, since most often the points shared are between faces
    // recently added).
    int v = vertices.count() - 1;
    for ( ; v >= from; --v)
        if (qFuzzyCompare(vertices[v], a))
            break;
    if (v < from)
    {
        // The vert was not found so add it as a new one, and initialize
        // its corresponding normal
        v = vertices.count();
        vertices.append(a);
        normals.append(n);
    }
    else
    {
        // Vert found, accumulate normals into corresponding normal slot.
        // Must call finalize once finished accumulating normals
        normals[v] += n;
    }
    // In both cases (found or not) reference the vert via its index
    faces.append(v);
}

void Geometry::appendFaceted(const QVector3D &a, const QVector3D &n)
{
    // Faceted normals are achieved by duplicating the vert for every
    // normal, so that faces meeting at a vert get a sharp edge.
    int v = vertices.count();
    vertices.append(a);
    normals.append(n);
    faces.append(v);
}

Patch::Patch(Geometry *g)
   : start(g->faces.count())
   , count(0)
   , initv(g->vertices.count())
   , sm(Patch::Smooth)
   , geom(g)
{
    qSetColor(faceColor, QColor(Qt::darkGray));
}

void Patch::rotate(qreal deg, QVector3D axis)
{
    mat.rotate(deg, axis);
}

void Patch::translate(const QVector3D &t)
{
    mat.translate(t);
}

static inline void qMultMatrix(const QMatrix4x4 &mat)
{
    if (sizeof(qreal) == sizeof(GLfloat))
        glMultMatrixf((GLfloat*)mat.constData());
#ifndef QT_OPENGL_ES
    else if (sizeof(qreal) == sizeof(GLdouble))
        glMultMatrixd((GLdouble*)mat.constData());
#endif
    else
    {
        GLfloat fmat[16];
        qreal const *r = mat.constData();
        for (int i = 0; i < 16; ++i)
            fmat[i] = r[i];
        glMultMatrixf(fmat);
    }
}

//! [2]
void Patch::draw() const
{
    glPushMatrix();
    qMultMatrix(mat);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, faceColor);

    const GLushort *indices = geom->faces.constData();
    glDrawElements(GL_TRIANGLES, count, GL_UNSIGNED_SHORT, indices + start);
    glPopMatrix();
}
//! [2]

void Patch::addTri(const QVector3D &a, const QVector3D &b, const QVector3D &c, const QVector3D &n)
{
    QVector3D norm = n.isNull() ? QVector3D::normal(a, b, c) : n;
    if (sm == Smooth)
    {
        geom->appendSmooth(a, norm, initv);
        geom->appendSmooth(b, norm, initv);
        geom->appendSmooth(c, norm, initv);
    }
    else
    {
        geom->appendFaceted(a, norm);
        geom->appendFaceted(b, norm);
        geom->appendFaceted(c, norm);
    }
    count += 3;
}

void Patch::addQuad(const QVector3D &a, const QVector3D &b,  const QVector3D &c, const QVector3D &d)
{
    QVector3D norm = QVector3D::normal(a, b, d);
    if (sm == Smooth)
    {
        addTri(a, b, c, norm);
        addTri(a, c, d, norm);
    }
    else
    {
        // If faceted share the two common verts
        addTri(a, b, c, norm);
        int k = geom->vertices.count();
        geom->appendSmooth(a, norm, k);
        geom->appendSmooth(c, norm, k);
        geom->appendFaceted(d, norm);
        count += 3;
    }
}

static inline QVector<QVector3D> extrude(const QVector<QVector3D> &verts, qreal depth)
{
    QVector<QVector3D> extr = verts;
    for (int v = 0; v < extr.count(); ++v)
        extr[v].setZ(extr[v].z() - depth);
    return extr;
}

class Rectoid
{
public:
    void translate(const QVector3D &t)
    {
        for (int i = 0; i < parts.count(); ++i)
            parts[i]->translate(t);
    }
    void rotate(qreal deg, QVector3D axis)
    {
        for (int i = 0; i < parts.count(); ++i)
            parts[i]->rotate(deg, axis);
    }

    // No special Rectoid destructor - the parts are fetched out of this member
    // variable, and destroyed by the new owner
    QList<Patch*> parts;
};


class RectTorus : public Rectoid
{
public:
    RectTorus(Geometry *g, qreal iRad, qreal oRad, qreal depth, int numSectors);
};

RectTorus::RectTorus(Geometry *g, qreal iRad, qreal oRad, qreal depth, int k)
{
    QVector<QVector3D> inside;
    QVector<QVector3D> outside;
    for (int i = 0; i < k; ++i) {
        qreal angle = (i * 2 * M_PI) / k;
        inside << QVector3D(iRad * qSin(angle), iRad * qCos(angle), depth / 2.0);
        outside << QVector3D(oRad * qSin(angle), oRad * qCos(angle), depth / 2.0);
    }
    inside << QVector3D(0.0, iRad, 0.0);
    outside << QVector3D(0.0, oRad, 0.0);
    QVector<QVector3D> in_back = extrude(inside, depth);
    QVector<QVector3D> out_back = extrude(outside, depth);

    // Create front, back and sides as separate patches so that smooth normals
    // are generated for the curving sides, but a faceted edge is created between
    // sides and front/back
    Patch *front = new Patch(g);
    for (int i = 0; i < k; ++i)
        front->addQuad(outside[i], inside[i],
                       inside[(i + 1) % k], outside[(i + 1) % k]);
    parts << front ;
}

QtLogo::QtLogo(const giac::gen &surfaceg, QObject *parent)
    : QObject(parent)
    , geom(new Geometry())
{
    //buildGeometry(divisions, scale);
          if (!giac::ckmatrix(surfaceg,true))
            return;

          //qDebug()<<"ckmatrix OK";
          qreal realiscmplx=giac::has_i(surfaceg),zmin=1e300,zmax=-1e300;
          giac::matrice & surface = *surfaceg._VECTptr;

          if (realiscmplx){
            if (find_zscale(surface,zmin,zmax))
                 realiscmplx=2*M_PI/(zmin-zmax);
          }

          //std::cout<<surface<<std::endl;

          Patch *surfpatch = new Patch(geom);
          surfpatch->setSmoothing(Patch::Smooth);



          int n=surface.size();
          //qDebug()<<"surface size "<<n;
          if (surfaceg.subtype==giac::_POLYEDRE__VECT){
             // implicit surface drawing with the given triangulation
              //qDebug()<<"surface _POLYEDRE__VECT";
             for (int i=0;i<n;++i){
               giac::vecteur & v=*surface[i]._VECTptr;
               giac::const_iterateur it=v.begin();
               qreal d1,d2,d3,e1,e2,e3,f1,f2,f3; // double in giac. PB with ARM?

                if (get_glvertex(*it,d1,d2,d3,realiscmplx,zmin,0) &&
                   get_glvertex(*(it+1),e1,e2,e3,realiscmplx,zmin,0) &&
                   get_glvertex(*(it+2),f1,f2,f3,realiscmplx,zmin,0)){
                   surfpatch->addTri(QVector3D(d1,d2,d3),QVector3D(e1,e2,e3),QVector3D(f1,f2,f3),QVector3D(0,0,0));
                   //qDebug()<<"creation de la surface";
                }
              }
          }
          else{
              /*if (surface.front()._VECTptr->size()<2){
                if (texture)
              glDisable(GL_TEXTURE_2D);
                return;
              }*/
              gen a,b,c,d;
              double xt=0,yt,dxt=double(1)/n,dyt;
              for (int i=1;i<n;++i,xt+=dxt){
                const vecteur & vprec=*surface[i-1]._VECTptr;
                const vecteur & v=*surface[i]._VECTptr;
                const_iterateur itprec=vprec.begin(),it=v.begin(),itend=v.end();
                yt=0; dyt=double(1)/(itend-it);
                a=*itprec; b=*it;
                double a1,a2,a3,b1,b2,b3,c1,c2,c3,d1,d2,d3;
                {
                get_glvertex(a,a1,a2,a3,realiscmplx,zmin,0);
                get_glvertex(b,b1,b2,b3,realiscmplx,zmin,0);
                ++it; ++itprec;
                for (;it!=itend;++it,++itprec,yt+=dyt){
                c = *itprec;
                d = *it;
                get_glvertex(c,c1,c2,c3,realiscmplx,zmin,0);
                get_glvertex(d,d1,d2,d3,realiscmplx,zmin,0);
                //qDebug()<<QVector3D(a1,a2,a3);
                QVector3D A(a1,a2,a3);
                QVector3D B(b1,b2,b3);
                QVector3D C(c1,c2,c3);
                QVector3D D(d1,d2,d3);
                //qDebug()<<A<<B<<C<<D;
                surfpatch->addQuad(A,C,D,B);
                //surfpatch->addTri(A,C,D,QVector3D::normal(A,C,B));
                //surfpatch->addTri(A,D,B,QVector3D::normal(A,C,B));
                a=c; a1=c1; a2=c2; a3=c3;
                b=d; b1=d1; b2=d2; b3=d3;
                }
                }
              }

          }





     parts << surfpatch;
    geom->finalize();
}

QtLogo::~QtLogo()
{
    qDeleteAll(parts);
    delete geom;
}

void QtLogo::setColor(QColor c)
{
    for (int i = 0; i < parts.count(); ++i)
        qSetColor(parts[i]->faceColor, c);
}

//! [3]
void QtLogo::buildGeometry(int divisions, qreal scale)
{
    qreal cw = cross_width * scale;
    qreal bt = bar_thickness * scale;
    qreal ld = logo_depth * scale;
    qreal th = tee_height *scale;

    RectTorus body(geom, 0.20, 0.30, 0.1, divisions);

    //parts << stem.parts << cross.parts << body.parts;
    parts << body.parts;

    geom->finalize();
}
//! [3]

//! [4]
void QtLogo::draw() const
{
    geom->loadArrays();

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);

    for (int i = 0; i < parts.count(); ++i)
        parts[i]->draw();

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
}
//! [4]
